@@ -1,35 +1,55 @@
 # -*- coding: utf-8 -*-
 
-# Copyright (c) 2006-2010 Filip Wasilewski <http://filipwasilewski.pl/>
+# Copyright (c) 2006-2012 Filip Wasilewski <http://en.ig.ma/>
 # See COPYING for license details.
 
-# $Id: wavelet_packets.py 154 2010-03-13 13:18:59Z filipw $
-
 """1D and 2D Wavelet packet transform module."""
 
+from __future__ import division, print_function, absolute_import
+
 __all__ = ["BaseNode", "Node", "WaveletPacket", "Node2D", "WaveletPacket2D"]
 
-import numerix
-from _pywt import Wavelet, dwt, idwt, dwt_max_level
-from multidim import dwt2, idwt2
+import numpy as np
+
+from ._pywt import Wavelet, dwt, idwt, dwt_max_level
+from .multidim import dwt2, idwt2
 
 
 def get_graycode_order(level, x='a', y='d'):
     graycode_order = [x, y]
-    for i in range(level-1):
+    for i in range(level - 1):
         graycode_order = [x + path for path in graycode_order] + \
                          [y + path for path in graycode_order[::-1]]
     return graycode_order
 
 
-class MustOverride(object):
-    def __init__(self, message): self.message= message
-    def __get__(self, obj, cls): raise NotImplementedError(self.message % {'cls': cls.__name__})
-
-
 class BaseNode(object):
-    PART_LEN = MustOverride("BaseNode.PART_LEN must be redefined in subclass %(cls)s.")
-    PARTS = MustOverride("BaseNode.PARTS must be redefined in subclass %(cls)s.")
+    """
+    BaseNode for wavelet packet 1D and 2D tree nodes.
+
+    The BaseNode is a base class for `Node` and `Node2D`.
+    It should not be used directly unless creating a new transformation
+    type. It is included here to document the common interface of 1D
+    and 2D node and wavelet packet transform classes.
+
+    Parameters
+    ----------
+    parent :
+        Parent node. If parent is None then the node is considered detached
+        (ie root).
+    data : 1D or 2D array
+        Data associated with the node. 1D or 2D numeric array, depending on the
+        transform type.
+    node_name :
+        A name identifying the coefficients type.
+        See `Node.node_name` and `Node2D.node_name`
+        for information on the accepted subnodes names.
+    """
+
+    # PART_LEN and PARTS attributes that define path tokens for node[] lookup
+    # must be defined in subclasses.
+    PART_LEN = None
+    PARTS = None
 
     def __init__(self, parent, data, node_name):
         self.parent = parent
@@ -83,14 +103,18 @@ class BaseNode(object):
         """
         Try to find the value of maximum decomposition level if it is not
         specified explicitly.
+
+        Parameters
+        ----------
+        evaluate_from : {'parent', 'subnodes'}
         """
         assert evaluate_from in ('parent', 'subnodes')
 
         if self._maxlevel is not None:
             return self._maxlevel
         elif self.data is not None:
-            return self.level + dwt_max_level(min(self.data.shape),
-                                              self.wavelet)
+            return self.level + dwt_max_level(
+                min(self.data.shape), self.wavelet)
 
         if evaluate_from == 'parent':
             if self.parent is not None:
@@ -104,6 +128,7 @@ class BaseNode(object):
                         return level
         return None
 
+    @property
     def maxlevel(self):
         if self._maxlevel is not None:
             return self._maxlevel
@@ -115,15 +140,23 @@ class BaseNode(object):
         if self._maxlevel is None:
             self._maxlevel = self._evaluate_maxlevel(evaluate_from='subnodes')
         return self._maxlevel
-    maxlevel = property(maxlevel)
 
+    @property
     def node_name(self):
         return self.path[-self.PART_LEN:]
-    node_name = property(node_name)
 
     def decompose(self):
         """
-        Decompose node data creating DWT coefficients subnodes."
+        Decompose node data creating DWT coefficients subnodes.
+
+        Performs Discrete Wavelet Transform on the `~BaseNode.data` and
+        returns transform coefficients.
+
+        Note
+        ----
+        Descends to subnodes and recursively
+        calls `~BaseNode.reconstruct` on them.
+
         """
         if self.level < self.maxlevel:
             return self._decompose()
@@ -136,8 +169,12 @@ class BaseNode(object):
     def reconstruct(self, update=False):
         """
         Reconstruct node from subnodes.
-        If update param is True, then reconstructed data replaces the current
-        node data.
+
+        Parameters
+        ----------
+        update : bool, optional
+            If True, then reconstructed data replaces the current
+            node data (default: False).
 
         Returns:
             - original node data if subnodes do not exist
@@ -148,15 +185,21 @@ class BaseNode(object):
         return self._reconstruct(update)
 
     def _reconstruct(self):
-        raise NotImplementedError() # override this in subclasses
+        raise NotImplementedError()  # override this in subclasses
 
     def get_subnode(self, part, decompose=True):
         """
-        Returns subnode.
+        Returns subnode or None (see `decomposition` flag description).
 
-        part      - subnode name
-        decompose - if True and subnode does not exist, it will be created using
-                    coefficients from DWT decomposition of the current node.
+        Parameters
+        ----------
+        part :
+            Subnode name
+        decompose : bool, optional
+            If the param is True and corresponding subnode does not
+            exist, the subnode will be created using coefficients
+            from the DWT decomposition of the current node.
+            (default: True)
         """
         self._validate_node_name(part)
         subnode = self._get_node(part)
@@ -169,14 +212,23 @@ class BaseNode(object):
         """
         Find node represented by the given path.
 
-        path - string composed of node names.
+        Similar to `~BaseNode.get_subnode` method with `decompose=True`, but
+        can access nodes on any level in the decomposition tree.
 
+        Parameters
+        ----------
+        path : str
+            String composed of node names. See `Node.node_name` and
+            `Node2D.node_name` for node naming convention.
+
+        Notes
+        -----
         If node does not exist yet, it will be created by decomposition of its
         parent node.
         """
-        if isinstance(path, basestring):
+        if isinstance(path, str):
             if (self.maxlevel is not None
-                and len(path) > self.maxlevel * self.PART_LEN):
+                    and len(path) > self.maxlevel * self.PART_LEN):
                 raise IndexError("Path length is out of range.")
             if path:
                 return self.get_subnode(path[0:self.PART_LEN], True)[
@@ -189,15 +241,21 @@ class BaseNode(object):
 
     def __setitem__(self, path, data):
         """
-        Set node represented by the given path with a new value.
+        Set node or node's data in the decomposition tree. Nodes are
+        identified by string `path`.
 
-        path - string composed of node names.
-        data - array or BaseNode subclass.
+        Parameters
+        ----------
+        path : str
+            String composed of node names.
+        data : array or BaseNode subclass.
         """
 
-        if isinstance(path, basestring):
-            if (self.maxlevel is not None
-                and len(self.path) + len(path) > self.maxlevel * self.PART_LEN):
+        if isinstance(path, str):
+            if (
+                self.maxlevel is not None
+                and len(self.path) + len(path) > self.maxlevel * self.PART_LEN
+            ):
                 raise IndexError("Path length out of range.")
             if path:
                 subnode = self.get_subnode(path[0:self.PART_LEN], False)
@@ -207,9 +265,9 @@ class BaseNode(object):
                 subnode[path[self.PART_LEN:]] = data
             else:
                 if isinstance(data, BaseNode):
-                    self.data = numerix.as_float_array(data.data)
+                    self.data = np.asarray(data.data, dtype=np.float64)
                 else:
-                    self.data = numerix.as_float_array(data)
+                    self.data = np.asarray(data, dtype=np.float64)
         else:
             raise TypeError("Invalid path parameter type - expected string but"
                             " got %s." % type(path))
@@ -217,13 +275,19 @@ class BaseNode(object):
     def __delitem__(self, path):
         """
         Remove node from the tree.
+
+        Parameters
+        ----------
+        path : str
+            String composed of node names.
         """
         node = self[path]
-        # don't clear node value and subnodes (node may still exist outside the tree)
-        ## node._init_subnodes()
-        ## node.data = None
+        # don't clear node value and subnodes (node may still exist outside
+        # the tree)
+        # # node._init_subnodes()
+        # # node.data = None
         parent = node.parent
-        node.parent = None # TODO
+        node.parent = None  # TODO
         if parent and node.node_name:
             parent._delete_node(node.node_name)
 
@@ -233,7 +297,7 @@ class BaseNode(object):
 
     def has_any_subnode(self):
         for part in self.PARTS:
-            if self._get_node(part) is not None: # and not .is_empty
+            if self._get_node(part) is not None:  # and not .is_empty
                 return True
         return False
     has_any_subnode = property(has_any_subnode)
@@ -241,8 +305,14 @@ class BaseNode(object):
     def get_leaf_nodes(self, decompose=False):
         """
         Returns leaf nodes.
+
+        Parameters
+        ----------
+        decompose : bool, optional
+            (default: True)
         """
         result = []
+
         def collect(node):
             if node.level == node.maxlevel and not node.is_empty:
                 result.append(node)
@@ -254,30 +324,52 @@ class BaseNode(object):
         self.walk(collect, decompose=decompose)
         return result
 
-    def walk(self, func, args=(), kwargs={}, decompose=True):
+    def walk(self, func, args=(), kwargs=None, decompose=True):
         """
-        Walk tree and call func on every node -> func(node, *args)
-        If func returns True, descending to subnodes will continue.
+        Traverses the decomposition tree and calls
+        ``func(node, *args, **kwargs)`` on every node. If `func` returns True,
+        descending to subnodes will continue.
 
-        func - callable
-        args - func parms
-        kwargs - func keyword params
+        Parameters
+        ----------
+        func : callable
+            Callable accepting `BaseNode` as the first param and
+            optional positional and keyword arguments
+        args :
+            func params
+        kwargs :
+            func keyword params
+        decompose : bool, optional
+            If True (default), the method will also try to decompose the tree
+            up to the `maximum level <BaseNode.maxlevel>`.
         """
+        if kwargs is None:
+            kwargs = {}
         if func(self, *args, **kwargs) and self.level < self.maxlevel:
             for part in self.PARTS:
                 subnode = self.get_subnode(part, decompose)
                 if subnode is not None:
                     subnode.walk(func, args, kwargs, decompose)
 
-    def walk_depth(self, func, args=(), kwargs={}, decompose=False):
+    def walk_depth(self, func, args=(), kwargs=None, decompose=False):
         """
         Walk tree and call func on every node starting from the bottom-most
         nodes.
 
-        func - callable
-        args - func parms
-        kwargs - func keyword params
+        Parameters
+        ----------
+        func : callable
+            Callable accepting :class:`BaseNode` as the first param and
+            optional positional and keyword arguments
+        args :
+            func params
+        kwargs :
+            func keyword params
+        decompose : bool, optional
+            (default: False)
         """
+        if kwargs is None:
+            kwargs = {}
         if self.level < self.maxlevel:
             for part in self.PARTS:
                 subnode = self.get_subnode(part, decompose)
@@ -293,7 +385,7 @@ class Node(BaseNode):
     """
     WaveletPacket tree node.
 
-    Subnodes are called ``a`` and ``d``, just like approximation
+    Subnodes are called `a` and `d`, just like approximation
     and detail coefficients in the Discrete Wavelet Transform.
     """
 
@@ -303,15 +395,22 @@ class Node(BaseNode):
     PART_LEN = 1
 
     def _create_subnode(self, part, data=None, overwrite=True):
-        return self._create_subnode_base(node_cls=Node, part=part, data=data, overwrite=overwrite)
+        return self._create_subnode_base(node_cls=Node, part=part, data=data,
+                                         overwrite=overwrite)
 
     def _decompose(self):
+        """
+
+        See also
+        --------
+        dwt : for 1D Discrete Wavelet Transform output coefficients.
+        """
         if self.is_empty:
             data_a, data_d = None, None
             if self._get_node(self.A) is None:
                 self._create_subnode(self.A, data_a)
-            if self._get_node(self.B) is None:
-                self._create_subnode(self.B, data_b)
+            if self._get_node(self.D) is None:
+                self._create_subnode(self.D, data_d)
         else:
             data_a, data_d = dwt(self.data, self.wavelet, self.mode)
             self._create_subnode(self.A, data_a)
@@ -323,9 +422,9 @@ class Node(BaseNode):
         node_a, node_d = self._get_node(self.A), self._get_node(self.D)
 
         if node_a is not None:
-            data_a = node_a.reconstruct() # TODO: (update) ???
+            data_a = node_a.reconstruct()  # TODO: (update) ???
         if node_d is not None:
-            data_d = node_d.reconstruct() # TODO: (update) ???
+            data_d = node_d.reconstruct()  # TODO: (update) ???
 
         if data_a is None and data_d is None:
             raise ValueError("Node is a leaf node and cannot be reconstructed"
@@ -341,8 +440,9 @@ class Node(BaseNode):
 class Node2D(BaseNode):
     """
     WaveletPacket tree node.
-    Subnodes are called 'a' (LL), 'h' (LH), 'v' (HL) and  'd' (HH), like approximation and
-    detail coefficients in 2D Discrete Wavelet Transform
+
+    Subnodes are called 'a' (LL), 'h' (LH), 'v' (HL) and  'd' (HH), like
+    approximation and detail coefficients in the 2D Discrete Wavelet Transform
     """
 
     LL = 'a'
@@ -354,33 +454,49 @@ class Node2D(BaseNode):
     PART_LEN = 1
 
     def _create_subnode(self, part, data=None, overwrite=True):
-        return self._create_subnode_base(node_cls=Node2D, part=part, data=data, overwrite=overwrite)
+        return self._create_subnode_base(node_cls=Node2D, part=part, data=data,
+                                         overwrite=overwrite)
 
     def _decompose(self):
+        """
+        See also
+        --------
+        dwt2 : for 2D Discrete Wavelet Transform output coefficients.
+        """
         if self.is_empty:
             data_ll, data_lh, data_hl, data_hh = None, None, None, None
         else:
-            data_ll, (data_lh, data_hl, data_hh) = dwt2(self.data, self.wavelet, self.mode)
+            data_ll, (data_lh, data_hl, data_hh) =\
+                dwt2(self.data, self.wavelet, self.mode)
         self._create_subnode(self.LL, data_ll)
         self._create_subnode(self.LH, data_lh)
         self._create_subnode(self.HL, data_hl)
         self._create_subnode(self.HH, data_hh)
-        return self._get_node(self.LL), self._get_node(self.LH), self._get_node(self.HL), self._get_node(self.HH)
+        return (self._get_node(self.LL), self._get_node(self.LH),
+                self._get_node(self.HL), self._get_node(self.HH))
 
     def _reconstruct(self, update):
         data_ll, data_lh, data_hl, data_hh = None, None, None, None
 
-        node_ll, node_lh, node_hl, node_hh = \
-            self._get_node(self.LL), self._get_node(self.LH), self._get_node(self.HL), self._get_node(self.HH)
+        node_ll, node_lh, node_hl, node_hh =\
+            self._get_node(self.LL), self._get_node(self.LH),\
+            self._get_node(self.HL), self._get_node(self.HH)
 
-        if node_ll is not None: data_ll = node_ll.reconstruct()
-        if node_lh is not None: data_lh = node_lh.reconstruct()
-        if node_hl is not None: data_hl = node_hl.reconstruct()
-        if node_hh is not None: data_hh = node_hh.reconstruct()
+        if node_ll is not None:
+            data_ll = node_ll.reconstruct()
+        if node_lh is not None:
+            data_lh = node_lh.reconstruct()
+        if node_hl is not None:
+            data_hl = node_hl.reconstruct()
+        if node_hh is not None:
+            data_hh = node_hh.reconstruct()
 
         if (data_ll is None and data_lh is None
-            and data_hl is None and data_hh is None):
-            raise ValueError("Tree is missing data - all subnodes of `%s` node are None. Cannot reconstruct node." % self.path)
+                and data_hl is None and data_hh is None):
+            raise ValueError(
+                "Tree is missing data - all subnodes of `%s` node "
+                "are None. Cannot reconstruct node." % self.path
+            )
         else:
             coeffs = data_ll, (data_lh, data_hl, data_hh)
             rec = idwt2(coeffs, self.wavelet, self.mode)
@@ -403,11 +519,19 @@ class WaveletPacket(Node):
     """
     Data structure representing Wavelet Packet decomposition of signal.
 
-    data     - original data (signal)
-    wavelet  - wavelet used in DWT decomposition and reconstruction
-    mode     - signal extension mode - see MODES
-    maxlevel - maximum level of decomposition (will be computed if not
-               specified)
+    Parameters
+    ----------
+    data : 1D ndarray
+        Original data (signal)
+    wavelet : Wavelet object or name string
+        Wavelet used in DWT decomposition and reconstruction
+    mode : str, optional
+        Signal extension mode for the `dwt` and `idwt` decomposition and
+        reconstruction functions.
+    maxlevel : int, optional
+        Maximum level of decomposition.
+        If None, it will be calculated based on the `wavelet` and `data`
+        length using `pywt.dwt_max_level`.
     """
     def __init__(self, data, wavelet, mode='sym', maxlevel=None):
         super(WaveletPacket, self).__init__(None, data, "")
@@ -418,8 +542,8 @@ class WaveletPacket(Node):
         self.mode = mode
 
         if data is not None:
-            data = numerix.as_float_array(data)
-            assert len(data.shape) == 1
+            data = np.asarray(data, dtype=np.float64)
+            assert data.ndim == 1
             self.data_size = data.shape[0]
             if maxlevel is None:
                 maxlevel = dwt_max_level(self.data_size, self.wavelet)
@@ -432,8 +556,11 @@ class WaveletPacket(Node):
         """
         Reconstruct data value using coefficients from subnodes.
 
-        If update is True, then data values will be replaced by
-        reconstruction values, also in subnodes.
+        Parameters
+        ----------
+        update : bool, optional
+            If True (default), then data values will be replaced by
+            reconstruction values, also in subnodes.
         """
         if self.has_any_subnode:
             data = super(WaveletPacket, self).reconstruct(update)
@@ -442,14 +569,29 @@ class WaveletPacket(Node):
             if update:
                 self.data = data
             return data
-        return self.data # return original data
+        return self.data  # return original data
 
     def get_level(self, level, order="natural", decompose=True):
         """
         Returns all nodes on the specified level.
 
-        order - "natural" - left to right in tree
-              - "freq" - band ordered
+        Parameters
+        ----------
+        level : int
+            Specifies decomposition `level` from which the nodes will be
+            collected.
+        order : {'natural', 'freq'}, optional
+            - "natural" - left to right in tree (default)
+            - "freq" - band ordered
+        decompose : bool, optional
+            If set then the method will try to decompose the data up
+            to the specified `level` (default: True).
+
+        Notes
+        -----
+        If nodes at the given level are missing (i.e. the tree is partially
+        decomposed) and the `decompose` is set to False, only existing nodes
+        will be returned.
         """
         assert order in ["natural", "freq"]
         if level > self.maxlevel:
@@ -479,11 +621,19 @@ class WaveletPacket2D(Node2D):
     """
     Data structure representing 2D Wavelet Packet decomposition of signal.
 
-    data     - original data (signal)
-    wavelet  - wavelet used in DWT decomposition and reconstruction
-    mode     - signal extension mode - see MODES
-    maxlevel - maximum level of decomposition (will be computed if not
-               specified)
+    Parameters
+    ----------
+    data : 2D ndarray
+        Data associated with the node.
+    wavelet : Wavelet object or name string
+        Wavelet used in DWT decomposition and reconstruction
+    mode : str, optional
+        Signal extension mode for the `dwt` and `idwt` decomposition and
+        reconstruction functions.
+    maxlevel : int
+        Maximum level of decomposition.
+        If None, it will be calculated based on the `wavelet` and `data`
+        length using `pywt.dwt_max_level`.
     """
     def __init__(self, data, wavelet, mode='sp1', maxlevel=None):
         super(WaveletPacket2D, self).__init__(None, data, "")
@@ -494,8 +644,8 @@ class WaveletPacket2D(Node2D):
         self.mode = mode
 
         if data is not None:
-            data = numerix.as_float_array(data)
-            assert len(data.shape) == 2
+            data = np.asarray(data, dtype=np.float64)
+            assert data.ndim == 2
             self.data_size = data.shape
             if maxlevel is None:
                 maxlevel = dwt_max_level(min(self.data_size), self.wavelet)
@@ -507,8 +657,11 @@ class WaveletPacket2D(Node2D):
         """
         Reconstruct data using coefficients from subnodes.
 
-        If update is set to True then the coefficients of the current node
-        and its subnodes will be replaced with values from reconstruction.
+        Parameters
+        ----------
+        update : bool, optional
+            If True (default) then the coefficients of the current node
+            and its subnodes will be replaced with values from reconstruction.
         """
         if self.has_any_subnode:
             data = super(WaveletPacket2D, self).reconstruct(update)
@@ -517,17 +670,25 @@ class WaveletPacket2D(Node2D):
             if update:
                 self.data = data
             return data
-        return self.data # return original data
+        return self.data  # return original data
 
     def get_level(self, level, order="natural", decompose=True):
         """
         Returns all nodes from specified level.
-        
-        If order is `natural`, a flat list is returned.
-        
-        If order is `freq`, a 2d structure with rows and cols
-        sorted by corresponding dimension frequency of 2d
-        coefficient array (adapted from 1d case).
+
+        Parameters
+        ----------
+        level : int
+            Decomposition `level` from which the nodes will be
+            collected.
+        order : {'natural', 'freq'}, optional
+            If `natural` (default) a flat list is returned.
+            If `freq`, a 2d structure with rows and cols
+            sorted by corresponding dimension frequency of 2d
+            coefficient array (adapted from 1d case).
+        decompose : bool, optional
+            If set then the method will try to decompose the data up
+            to the specified `level` (default: True).
         """
         assert order in ["natural", "freq"]
         if level > self.maxlevel:
@@ -543,11 +704,12 @@ class WaveletPacket2D(Node2D):
             return True
 
         self.walk(collect, decompose=decompose)
-        
+
         if order == "freq":
             nodes = {}
-            for (row_path, col_path), node in [(self.expand_2d_path(node.path), node)
-                                               for node in result]:
+            for (row_path, col_path), node in [
+                (self.expand_2d_path(node.path), node) for node in result
+            ]:
                 nodes.setdefault(row_path, {})[col_path] = node
             graycode_order = get_graycode_order(level, x='l', y='h')
             nodes = [nodes[path] for path in graycode_order if path in nodes]
@@ -557,4 +719,3 @@ class WaveletPacket2D(Node2D):
                     [row[path] for path in graycode_order if path in row]
                 )
         return result
-
